# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Source file module."""
import os
import shutil
import stat
from pathlib import Path

from mindinsight.wizard.common.exceptions import OSPermissionError, TemplateFileError


class SourceFile:
    """Network code generator."""
    file_relative_path = ''
    template_file_path = ''
    content = ''

    @staticmethod
    def _make_dir(directory):
        permissions = os.R_OK | os.W_OK | os.X_OK
        mode = permissions << 6
        os.makedirs(directory, mode=mode, exist_ok=True)
        return directory

    def write(self, project_directory):
        """Generate the network files."""
        template_file_path = Path(self.template_file_path)
        if not template_file_path.is_file():
            raise TemplateFileError("Template file %s is not exist." % self.template_file_path)
        new_file_path = os.path.join(project_directory, self.file_relative_path)
        self._make_dir(os.path.dirname(new_file_path))
        with open(new_file_path, 'w', encoding='utf-8') as fp:
            fp.write(self.content)
        try:
            shutil.copymode(self.template_file_path, new_file_path)
            os.chmod(new_file_path, stat.S_IRUSR | stat.S_IWUSR)
            self.set_writeable(new_file_path)
            if new_file_path.endswith('.sh'):
                self.set_executable(new_file_path)
        except OSError:
            raise OSPermissionError("Notice: Set permission bits failed on %s." % new_file_path)

    @staticmethod
    def set_writeable(file_name):
        """Add write permission."""
        if not os.access(file_name, os.W_OK):
            file_stat = os.stat(file_name)
            permissions = stat.S_IMODE(file_stat.st_mode) | stat.S_IWUSR
            os.chmod(file_name, permissions)

    @staticmethod
    def set_executable(file_name):
        """Add executable permission."""
        if not os.access(file_name, os.X_OK):
            file_stat = os.stat(file_name)
            permissions = stat.S_IMODE(file_stat.st_mode) | stat.S_IXUSR
            os.chmod(file_name, permissions)



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Web service entrance."""
import os
import re
import shlex
import stat
import subprocess
import sys
import time
from enum import Enum, unique

from gunicorn.glogging import Logger
from mindinsight.backend.config import WEB_CONFIG_DIR
from mindinsight.conf import settings, get_web_address
from mindinsight.utils.log import setup_logger

MINDBOARD_APP_MODULE = "mindinsight.backend.application:APP"
GUNICORN_LOGGER = "mindinsight.backend.run.GunicornLogger"

_MIN_PORT = 1
_MAX_PORT = 65535


@unique
class ServerStateEnum(Enum):
    """
    The service startup status are as follows: "unknown", "failed" and "success"
    """
    UNKNOWN = "unknown"
    FAILED = "failed"
    SUCCESS = "success"


def _get_file_size(file_path):
    """
    Get the file size.

    Args:
        file_path (str): The file path.

    Returns:
        int, the file size. If file is not existed, then return 0.
    """
    try:
        file_size = os.path.getsize(file_path)
    except FileNotFoundError:
        file_size = 0
    return file_size


def _is_match_one(sub_string_list, src_string):
    """
    Whether the sub-string in the list can match the source string.

    Args:
        sub_string_list (list): The sub-string list.
        src_string (str): The source string.

    Returns:
        bool, if matched return True, else return False.
    """
    for match_info in sub_string_list:
        if match_info in src_string:
            return True
    return False


def _check_stat_from_log(pid, log_info):
    """
    Determine the service startup status based on the log information.

    Args:
        pid (int): The gunicorn process ID.
        log_info (str): The output log of service startup.

    Returns:
        str, the state value that is one of the follows: "unknown", "failed" and "success".
    """
    server_state = ServerStateEnum.UNKNOWN.value

    # should be synchronized to startup log in gunicorn post_worker_init hook
    # refer to mindinsight/backend/config/gunicorn_conf.py
    match_success_info = "Server pid: %d, start to listening." % pid

    common_failed_info_list = [
        "[ERROR] Retrying in 1 second",
        "[INFO] Reason: App failed to load",
        "[ERROR] Exception in worker process"
    ]
    re_pattern = "\\[ERROR\\].+%s.+%d" % \
                 (settings.HOST, int(settings.PORT))

    # matched failed output log by fuzzy match
    if re.search(re_pattern, log_info) or \
            _is_match_one(common_failed_info_list, log_info):
        server_state = ServerStateEnum.FAILED.value

    if match_success_info in log_info:
        server_state = ServerStateEnum.SUCCESS.value

    return server_state


def _get_error_log_path():
    """
    Get gunicorn error log path.

    Returns:
        str, the path of error log.
    """

    path = os.path.join(settings.WORKSPACE, 'log/gunicorn/error.{}.log'.format(settings.PORT))
    errorlog_abspath = os.path.realpath(path)
    return errorlog_abspath


def _get_access_log_path():
    """Get gunicorn access log path."""
    access_log_path = os.path.join(settings.WORKSPACE, 'log/gunicorn/access.{}.log'.format(settings.PORT))
    access_log_path = os.path.realpath(access_log_path)
    return access_log_path


def _check_state_from_log(pid, log_abspath, start_pos=0):
    """
    Check the service startup status based on the log file.

    Args:
        pid (int): The gunicorn process ID.
        log_abspath (str): Absolute path of the log file.
        start_pos (int): Offset position of the log file.

    Returns:
        dict, a dict with "state" and "prompt_message" key.
        The value of the "state" key is as follows:"unknown", "failed" and "success".
        The value of the "prompt_message" key is a list of prompt messages.

    """
    state_result = {"state": ServerStateEnum.UNKNOWN.value, "prompt_message": []}
    prompt_messages = []
    with open(log_abspath) as f_log:
        f_log.seek(start_pos)
        for line in f_log.readlines():
            log_result = _check_stat_from_log(pid, line)
            # ignore "unknown" result
            if log_result != ServerStateEnum.UNKNOWN.value:
                state_result["state"] = log_result

            if log_result == ServerStateEnum.FAILED.value:
                prompt_messages.append(line.strip())
                prompt_messages.append(
                    "more failed details in log: %s" % log_abspath)
                break
    if log_result == ServerStateEnum.UNKNOWN.value:
        prompt_messages.append(
            "more details in log: %s" % log_abspath)
    state_result["prompt_message"].append(
        "service start state: %s" % state_result["state"])
    for prompt_message in prompt_messages:
        state_result["prompt_message"].append(prompt_message)

    return state_result


def _check_server_start_stat(pid, log_abspath, log_pos):
    """
    Checking the Server Startup Status.

    Args:
        pid (int): The gunicorn process ID.
        log_abspath (str): The log file path.

    Returns:
        dict, an dict object that contains the state and prompt_message fields.
        The state values are as follows: "unknown", "failed" and "success".

    """
    state_result = {"state": ServerStateEnum.UNKNOWN.value, "prompt_message": []}
    # return unknown when not config gunicorn error log file
    if not log_abspath:
        return state_result

    # sleep 1 second for gunicorn master to be ready
    time.sleep(1)

    try_cnt = 10
    while try_cnt > 0:
        try_cnt -= 1
        time.sleep(1)
        file_size = _get_file_size(log_abspath)
        if file_size > log_pos:
            state_result.update(_check_state_from_log(pid, log_abspath, log_pos))
            if state_result['state'] != ServerStateEnum.UNKNOWN.value:
                break

    if not state_result['prompt_message']:
        state_result["prompt_message"].append(
            "service start state: %s" % state_result["state"])

    return state_result


class GunicornLogger(Logger):
    """Rewrite gunicorn default logger."""

    def __init__(self, cfg):
        self.cfg = cfg
        self.access_log = setup_logger('gunicorn', 'access', formatter='%(message)s')
        self.error_log = setup_logger('gunicorn', 'error', formatter=self.error_fmt)
        access_log_path = _get_access_log_path()
        error_log_path = _get_error_log_path()
        os.chmod(access_log_path, stat.S_IREAD | stat.S_IWRITE)
        os.chmod(error_log_path, stat.S_IREAD | stat.S_IWRITE)
        super(GunicornLogger, self).__init__(cfg)

    def now(self):
        """Get log format."""
        return time.strftime('[%Y-%m-%d-%H:%M:%S %z]')

    def setup(self, cfg):
        """Rewrite the setup method of Logger, and we don't need to do anything"""


def start():
    """Start web service."""
    gunicorn_conf_file = os.path.join(WEB_CONFIG_DIR, "gunicorn_conf.py")
    cmd = "gunicorn " \
          "-b {host}:{port} {app_module} " \
          "-c {conf_file} " \
          "--logger-class {logger_class} " \
          "--access-logformat {log_format}"\
        .format(host=settings.HOST,
                port=settings.PORT,
                conf_file=gunicorn_conf_file,
                app_module=MINDBOARD_APP_MODULE,
                logger_class=GUNICORN_LOGGER,
                log_format=settings.GUNICORN_ACCESS_FORMAT
                )

    error_log_abspath = _get_error_log_path()

    # Init the logger file
    setup_logger('gunicorn', 'error')
    log_handler = open(error_log_abspath, 'a+')
    pre_log_pos = _get_file_size(error_log_abspath)
    # start server
    process = subprocess.Popen(
        shlex.split(cmd),
        shell=False,
        # Change stdout to DEVNULL to prevent broken pipe error when creating new processes.
        stdin=subprocess.DEVNULL,
        stdout=log_handler,
        stderr=subprocess.STDOUT
    )

    # sleep 1 second for gunicorn application to load modules
    time.sleep(1)

    # check if gunicorn application is running
    console = setup_logger('mindinsight', 'console', console=True, logfile=False, formatter='%(message)s')
    if process.poll() is not None:
        console.error("Start MindInsight failed. See log for details, log path: %s.", error_log_abspath)
        sys.exit(1)
    else:
        state_result = _check_server_start_stat(process.pid, error_log_abspath, pre_log_pos)
        # print gunicorn start state to stdout
        web_address = get_web_address()
        console.info(web_address)
        for line in state_result["prompt_message"]:
            console.info(line)
        if state_result["state"] == ServerStateEnum.FAILED.value:
            sys.exit(1)


if __name__ == '__main__':
    start()



# Copyright 2020-2023 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The Timeline Analyser."""
import csv
import json
import os
import glob
import re
import time
from decimal import Decimal
from concurrent.futures import ThreadPoolExecutor
from marshmallow import ValidationError

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.log import logger


def get_diff_time(rank_id, prof_path):
    """
    Get difference time between ranks
    """
    profiler_info_file = os.path.join(prof_path, os.pardir, f'profiler_info_{rank_id}.json')
    if not os.path.exists(profiler_info_file):
        return Decimal(0).quantize(Decimal('0.000'))

    with open(profiler_info_file, 'r') as fr:
        diff_time = json.load(fr).get('diff_time', 0)

    return Decimal(diff_time).quantize(Decimal('0.000'))


def get_rank_id_from_info_json(pro_path):
    """
    Get rank id and device id from PROFXXX
    """
    info_json = ""
    rank_id = 0
    device_id = 0
    for root, _, files in os.walk(pro_path):
        for file in files:
            if "info.json." in file and ".done" not in file:
                info_json = os.path.join(root, file)
                break

    if info_json:
        with open(info_json, "r+") as f:
            info = json.load(f)
        rank_id = int(info.get('rank_id', rank_id))
        device_id = int(info.get('devices', device_id))

    if rank_id == -1:
        rank_id = 0

    return rank_id, device_id


def get_timeline_info(prof_dirs):
    """
    Get the newest prof dirs
    """
    # need get the newest dirs
    timeline_info = {}

    for prof_path in prof_dirs:
        rank_id, _ = get_rank_id_from_info_json(prof_path)
        ts_difference_us = get_diff_time(rank_id, prof_path)
        if rank_id is None:
            logger.warning('Could not find the rank id in %s, ignore this file.', prof_path)
            continue

        if rank_id not in timeline_info or (rank_id in timeline_info and prof_path > timeline_info.get(rank_id)[0]):
            prof_path = os.path.join(prof_path, 'mindstudio_profiler_output')
            timeline_info[rank_id] = (prof_path, ts_difference_us)

    return timeline_info


def get_job_dir(parent_path):
    job_path_list = glob.glob(fr'{parent_path}/PROF_*_*')
    return get_timeline_info(job_path_list)


def get_newest_file(file_list):
    '''
    Find the newest files
    :param file_list:
    :return:
    '''
    newest_file_list = []
    newest_timestamp = '0'
    for file_path in file_list:
        timestamp = file_path.split('.')[0].split('/')[-1].split('_')[-1]
        newest_timestamp = max(timestamp, newest_timestamp)

    for file_path in file_list:
        if file_path.split('.')[0].split('/')[-1].split('_')[-1] == newest_timestamp:
            newest_file_list.append(file_path)

    newest_file_list.sort()
    return newest_file_list


class MsprofTimelineAnalyser(BaseAnalyser):
    """
    Analyse timeline data from file.
    """

    def __init__(self, profiling_dir, device_id=None):
        super(MsprofTimelineAnalyser, self).__init__(profiling_dir, device_id)
        self.top_scope_name = ('Default', 'Gradients', 'recompute_Default')
        self.step_trace_index = 1
        self.cann_index = 2
        self.scope_index = 3
        self.ascend_hardware_index = 4
        self.hccl_index = 5
        self.cpu_index = 6
        self.overlap_index = 7

    def get_merged_timeline(self, rank_list, model_list, kind, merge_model=True, scope_name=False):
        """
        Get the merged timeline
        """

        # get all job path, like PROF_*
        sub_dirs = get_job_dir(self._profiling_dir)

        if rank_list:
            new_sub_dirs = {}
            for key, value in sub_dirs.items():
                if key in rank_list:
                    new_sub_dirs[key] = value
            sub_dirs = new_sub_dirs

        if not sub_dirs:
            logger.error('Could not found any rank from %s', rank_list)
            return []

        if kind == 'summary':
            start = time.time()
            summary_data = self._get_summary_timeline_data(sub_dirs, merge_model)
            logger.info("Summary timeline time consuming: %s", time.time() - start)
            return summary_data

        if kind == 'detail':
            start = time.time()
            detail_data = self._get_detail_timeline_data(sub_dirs, model_list, merge_model, scope_name)
            logger.info("Detail timeline time consuming: %s", time.time() - start)
            return detail_data
        return []

    def parse_cpu_timeline(self, file_list, rank_id, difference_ts, scope_name):
        """Load cpu operator data from file"""
        ms_to_us = 1e3
        ps_to_ns = 1e-3
        new_pid = int(f'{self.cpu_index}{rank_id}')
        process_list = [{"name": "process_name",
                         "pid": new_pid,
                         "args": {
                             "name": f"CPU OP Rank{rank_id}"
                         },
                         "ph": "M"
                         }, {"name": "process_sort_index", "pid": new_pid,
                             "args": {"sort_index": self.cpu_index}, "ph": "M"}
                        ]
        tid_set = set()
        thread_list = []
        new_timeline = []
        scope_data = []
        try:
            flags = os.O_RDONLY
            for file_path in file_list:
                with os.fdopen(os.open(file_path, flags, 0o400), 'r') as fr:
                    for line in fr:
                        op_list = line.strip().split(';')
                        op_full_name = op_list[0]
                        time_arr = op_list[-1]
                        time_arr = time_arr.split(" ")
                        for time_str in time_arr:
                            ts, dur, tid = time_str.split(",")
                            ts = Decimal(ts).quantize(Decimal('0.000')) * Decimal(ps_to_ns).quantize(Decimal('0.000'))

                            if scope_name and op_full_name and op_full_name.startswith(self.top_scope_name):
                                te = ts + Decimal(dur).quantize(Decimal('0.000'))
                                scope_data.append((op_full_name.split('/')[:-1], ts, te))

                            ts -= difference_ts

                            if int(tid) not in tid_set:
                                tid_set.add(int(tid))
                                thread_list.append({"name": "thread_name",
                                                    "pid": new_pid,
                                                    "tid": int(tid),
                                                    "ph": "M",
                                                    'args': {'name': f'thread {tid}'}
                                                    })

                            new_timeline.append({'name': op_list[0],
                                                 'pid': new_pid,
                                                 'tid': int(tid),
                                                 'ph': 'X',
                                                 'ts': str(ts),
                                                 'dur': float(dur) * ms_to_us,
                                                 'args':
                                                     {'type': op_list[1]}
                                                 })
                break

            return process_list + thread_list + new_timeline, scope_data

        except (ValidationError, IOError, OSError, json.JSONDecodeError) as err:
            logger.error('parse_cann_data failed! please theck. detail: %s', err)
            return []

    def get_option(self):
        """
        Get the option values
        """
        # get all job path, like PROF_*
        sub_dirs = get_job_dir(self._profiling_dir)
        rank_list = list(sub_dirs.keys())
        rank_list.sort()

        _, model_merged = self._get_models(sub_dirs)
        model_list = list(model_merged)
        model_list.sort()

        return {'rank_list': rank_list, 'model_list': model_list}

    def _load(self):
        """Load data according to the parsed profiling files."""

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """

    def _parse_overlap_analysis_data(self, file_list, rank_id, difference_ts):
        """
        parse overlap analysis data
        """
        try:
            flags = os.O_RDONLY
            raw_data = []
            for file_path in file_list:
                with os.fdopen(os.open(file_path, flags, 0o400), 'r') as fr:
                    raw_data.extend(json.load(fr))

            pid = None
            for event in raw_data:
                if event.get('name') == 'process_name' and event.get("ph") == "M" and \
                        event.get('args').get('name') == 'Overlap Analysis':
                    pid = event.get('pid')
                    break

            if not pid:
                logger.warning('Could not found process_name pid. method: _parse_overlap_analysis_data')
                return []

            new_events = []
            new_pid = int(f'{self.overlap_index}{rank_id}')
            for event in raw_data:
                if event.get('pid') != pid:
                    continue

                if event.get('name') == 'process_name' and event.get("ph") == "M":
                    event["args"]["name"] += f" Rank{rank_id}"

                if event.get('name') == 'process_sort_index' and event.get("ph") == "M":
                    event["args"]["sort_index"] = self.overlap_index

                event['pid'] = new_pid
                if event.get('ts'):
                    ts = Decimal(event.get('ts')).quantize(Decimal('0.000'))
                    ts += difference_ts
                    event['ts'] = str(ts)

                new_events.append(event)

            return new_events

        except (ValidationError, IOError, OSError, json.JSONDecodeError) as err:
            logger.error('parse_overlap_analysis_data failed! please theck. detail: %s', err)
            return []

    def _parse_step_trace_metadata(self, raw_data, model_list):
        """
        Get step trace by merge models
        """
        pattern1 = re.compile(r'Step Trace\(Model ID:(\d)+\)')
        pattern2 = re.compile(r'(\d)+')
        tid_mapper = {}
        pid = None
        for event in raw_data:
            if event.get("ph") != "M":
                continue

            if event.get('name') == 'process_name':
                pid = event.get('pid')
                continue

            if event.get('name') == 'thread_name':
                arg_name = event.get('args', {}).get('name')
                arg_name = re.search(pattern1, arg_name)
                if not arg_name:
                    continue
                model_id = re.search(pattern2, arg_name.group())
                if not model_id:
                    continue
                model_id = model_id.group()
                tid = event.get('tid')
                if not model_list or int(model_id) in model_list:
                    tid_mapper[tid] = f'Model {model_id}'

        return tid_mapper, pid

    def _parse_step_trace_merge(self, old_pid, new_pid, rank_id, raw_data, tid_mapper, difference_ts):
        """merge step trace data"""
        new_events = [{
            "name": "process_name",
            "pid": new_pid,
            "args": {
                "name": f"Step Trace Rank{rank_id}"
            },
            "ph": "M"
        }, {
            "name": "process_sort_index",
            "pid": new_pid,
            "args": {
                "sort_index": self.step_trace_index
            },
            "ph": "M"
        }, {
            "name": "thread_name",
            "pid": new_pid,
            "tid": 0,
            "args": {
                "name": "iterations"
            },
            "ph": "M"
        }]

        for event in raw_data:
            arg_name = tid_mapper.get(event.get('tid'))
            if event.get('ph') == 'M' or event.get('pid') != old_pid or not arg_name:
                continue

            event_name = event.get('name').strip()
            if event.get('ph') == 'X' and event_name.startswith('Iteration') and len(
                    event_name.split(' ')) == 2:
                event['name'] = f"{arg_name} {event_name}"

            if event.get('ts'):
                ts = Decimal(event.get('ts')).quantize(Decimal('0.000'))
                ts -= difference_ts
                event['ts'] = str(ts)
            event['pid'] = new_pid
            event['tid'] = 0
            new_events.append(event)
        return new_events

    def _parse_step_trace_not_merge(self, old_pid, new_pid, rank_id, raw_data, difference_ts):
        """not merge step trace data"""
        new_events = []
        for event in raw_data:
            if event.get('pid') != old_pid:
                continue
            if event.get('name') == 'process_name' and event.get('ph') == 'M':
                event['args']['name'] = f"Step Trace Rank{rank_id}"
            elif event.get('name') == 'process_sort_index' and event.get('ph') == 'M':
                event['args']['sort_index'] = self.step_trace_index

            event['pid'] = new_pid
            if event.get('ts'):
                ts = Decimal(event.get('ts')).quantize(Decimal('0.000'))
                ts -= difference_ts
                event['ts'] = str(ts)
            new_events.append(event)
        return new_events

    def _parse_step_trace_data(self, file_list, rank_id, difference_ts, model_list, merge_model):
        """
        parse step trace data
        """
        try:
            flags = os.O_RDONLY
            raw_data = []
            for file_path in file_list:
                with os.fdopen(os.open(file_path, flags, 0o400), 'r') as fr:
                    raw_data.extend(json.load(fr))

            tid_mapper, pid = self._parse_step_trace_metadata(raw_data, model_list)
            if not pid:
                logger.error('Could not found process_name pid. method: _parse_step_trace_data')
                return []

            new_pid = int(f'{self.step_trace_index}{rank_id}')

            if merge_model:
                return self._parse_step_trace_merge(pid, new_pid, rank_id, raw_data, tid_mapper, difference_ts)

            return self._parse_step_trace_not_merge(pid, new_pid, rank_id, raw_data, difference_ts)

        except (ValidationError, IOError, OSError, json.JSONDecodeError) as err:
            logger.error('parse_step_trace_data failed! please theck. detail: %s', err)
            return []

    def _parse_msprof_metadata(self, new_pid_hardware, raw_data):
        """
        Get msprof by merge models
        """
        tid_mapper_hardware = {}
        pid_hardware = None
        pid_hccl = None
        pid_cann = None
        pid_overlap = None
        for event in raw_data:
            if event.get('name') == 'process_name' and event.get("ph") == "M" and \
                    event.get('args').get('name') == 'Ascend Hardware':
                pid_hardware = event.get('pid')

            elif event.get('name') == 'thread_name' and event.get("ph") == "M" and \
                    'Stream' in event.get('args').get('name'):
                event['pid'] = new_pid_hardware
                tid_mapper_hardware.update({event.get('tid'): event})

            elif event.get('name') == 'process_name' and event.get("ph") == "M" and \
                    event.get('args').get('name') == 'HCCL':
                pid_hccl = event.get('pid')

            elif event.get('name') == 'process_name' and event.get("ph") == "M" and \
                    event.get('args').get('name') == 'CANN':
                pid_cann = event.get('pid')

            elif event.get('name') == 'process_name' and event.get("ph") == "M" and \
                    event.get('args').get('name') == 'Overlap Analysis':
                pid_overlap = event.get('pid')

        result = (pid_hardware, tid_mapper_hardware, pid_hccl, pid_cann, pid_overlap)
        return result

    def _parse_msprof_raw_data(self, raw_data, difference_ts, tid_mapper_hardware, model_list, scope_name, **kwargs):
        """
        Parse the msprof raw data
        """

        new_events_hardware = []
        new_events_hccl = []
        new_events_cann = []
        new_events_overlap = []

        scope_data = []

        for event in raw_data:
            model_id = event.get('args', {}).get('Model Id')
            is_process = event.get('ph') == 'M' \
                         and (event.get('name') == 'process_name' or event.get('name') == 'process_sort_index')
            if is_process or (model_list and model_id not in model_list):
                continue

            op_full_name = event.get('name')
            if scope_name and op_full_name and op_full_name.startswith(self.top_scope_name):
                ts = Decimal(event.get('ts')).quantize(Decimal('0.000'))
                te = ts + Decimal(event.get('dur')).quantize(Decimal('0.000'))
                scope_data.append((op_full_name.split('/')[:-1], ts, te))

            if event.get('ts'):
                ts = Decimal(event.get('ts')).quantize(Decimal('0.000'))
                ts -= difference_ts
                event['ts'] = str(ts)

            if event.get('pid') == kwargs.get('pid_hardware') and event.get('ph') != 'M' \
                    and event.get('tid') in tid_mapper_hardware:
                event['pid'] = kwargs.get('new_pid_hardware')
                new_events_hardware.append(event)

            elif event.get('pid') == kwargs.get('pid_hccl'):
                event['pid'] = kwargs.get('new_pid_hccl')
                new_events_hccl.append(event)

            elif not model_list and event.get('pid') == kwargs.get('pid_cann'):
                event['pid'] = kwargs.get('new_pid_cann')
                new_events_cann.append(event)

            elif not model_list and event.get('pid') == kwargs.get('pid_overlap'):
                event['pid'] = kwargs.get('new_pid_overlap')
                new_events_overlap.append(event)

        return new_events_hardware + new_events_hccl + new_events_cann + new_events_overlap, scope_data

    def _parse_msprof_data(self, file_list, rank_id, difference_ts, model_list, scope_name):
        """
        parse ascend hardware and hccl and cann data
        """
        flags = os.O_RDONLY
        raw_data = []

        new_pid_hardware = int(f'{self.ascend_hardware_index}{rank_id}')
        new_pid_hccl = int(f'{self.hccl_index}{rank_id}')
        new_pid_cann = int(f'{self.cann_index}{rank_id}')
        new_pid_overlap = int(f'{self.overlap_index}{rank_id}')
        new_metadata = [
            {
                "name": "process_name",
                "pid": new_pid_hardware,
                "args": {
                    "name": f"Ascend Hardware Rank{rank_id}"
                },
                "ph": "M"
            }, {"name": "process_sort_index", "pid": new_pid_hardware,
                "args": {"sort_index": self.ascend_hardware_index}, "ph": "M"},
            {
                "name": "process_name",
                "pid": new_pid_hccl,
                "args": {
                    "name": f"HCCL Rank{rank_id}"
                },
                "ph": "M"
            }, {"name": "process_sort_index", "pid": new_pid_hccl,
                "args": {"sort_index": self.hccl_index}, "ph": "M"},
            {
                "name": "process_name",
                "pid": new_pid_cann,
                "args": {
                    "name": f"CANN Rank{rank_id}"
                },
                "ph": "M"
            }, {"name": "process_sort_index", "pid": new_pid_cann,
                "args": {"sort_index": self.cann_index}, "ph": "M"},
            {
                "name": "process_name",
                "pid": new_pid_overlap,
                "args": {
                    "name": f"Overlap Analysis Rank{rank_id}"
                },
                "ph": "M"
            }, {"name": "process_sort_index", "pid": new_pid_overlap,
                "args": {"sort_index": self.overlap_index}, "ph": "M"}
        ]

        try:
            for file_path in file_list:
                with os.fdopen(os.open(file_path, flags, 0o400), 'r') as fr:
                    raw_data.extend(json.load(fr))

            pid_hardware, tid_mapper_hardware, pid_hccl, pid_cann, pid_overlap \
                = self._parse_msprof_metadata(new_pid_hardware, raw_data)

            is_pid_valid = not pid_hardware and not pid_hccl and pid_cann and pid_overlap

            if is_pid_valid:
                logger.error('Could not found process_name pid. method: _parse_msprof_data')
                return []

            pid_dict = {'pid_hardware': pid_hardware, 'pid_hccl': pid_hccl,
                        'pid_cann': pid_cann, 'pid_overlap': pid_overlap,
                        'new_pid_hardware': new_pid_hardware, 'new_pid_hccl': new_pid_hccl,
                        'new_pid_cann': new_pid_cann, 'new_pid_overlap': new_pid_overlap}

            new_events, scope_data = self._parse_msprof_raw_data(raw_data, difference_ts, tid_mapper_hardware,
                                                                 model_list, scope_name, **pid_dict)
            return new_metadata + list(tid_mapper_hardware.values()) + new_events, scope_data

        except (ValidationError, IOError, OSError, json.JSONDecodeError) as err:
            logger.error('_parse_msprof_data failed! please theck. detail: %s', err)
            return []

    def _parse_scope_info(self, scope_data, rank_id, difference_ts):
        """parse scope layer"""
        if not scope_data:
            return []
        new_pid = int(f'{self.scope_index}{rank_id}')
        scope_data.sort(key=lambda x: x[1])
        process_list = [
            {"name": "process_name",
             "pid": new_pid,
             "args": {
                 "name": f"Scope Layer Rank{rank_id}"
             },
             "ph": "M"},
            {"name": "process_sort_index",
             "pid": new_pid,
             "args": {"sort_index": self.scope_index},
             "ph": "M"}
        ]

        new_events = []
        layer_stack = []
        for layer_name in scope_data[0][0]:
            layer_stack.append([layer_name, scope_data[0][1], scope_data[0][2]])

        for op in scope_data[1:]:
            if op[1] < layer_stack[0][2]:
                # 并行算子只保留前面的
                continue
            flag = True  # 判断上层是否合并， 上层不合并下层也不合并
            for layer_depth, layer_name in enumerate(op[0]):
                if layer_depth >= len(layer_stack):
                    layer_stack.append([layer_name, op[1], op[2]])
                else:
                    if layer_stack[layer_depth][0] == layer_name and flag:
                        layer_stack[layer_depth][2] = op[2]  # 合并
                    else:
                        ts = layer_stack[layer_depth][1]
                        ts -= difference_ts
                        new_events.append({
                            "name": layer_stack[layer_depth][0],
                            "pid": new_pid,
                            "tid": layer_depth,
                            "ph": "X",
                            "ts": str(ts),
                            "dur": float(layer_stack[layer_depth][2] - layer_stack[layer_depth][1])
                        })
                        layer_stack[layer_depth] = [layer_name, op[1], op[2]]
                        flag = False

        thread_list = []
        for index, layer in enumerate(layer_stack):
            thread_list.extend([{
                "name": "thread_name",
                "pid": new_pid,
                "tid": index,
                "args": {
                    "name": f"layer{index}"
                },
                "ph": "M"
            }, {
                "name": "thread_sort_index",
                "pid": new_pid,
                "tid": index,
                "args": {"sort_index": index},
                "ph": "M"
            }])
            if layer:
                ts = layer[1]
                ts -= difference_ts
                new_events.append({
                    "name": layer[0],
                    "pid": new_pid,
                    "tid": index,
                    "ph": "X",
                    "ts": str(ts),
                    "dur": float(layer[2] - layer[1])
                })

        return process_list + thread_list + new_events

    def _get_summary_timeline_data(self, sub_dirs, merge_model):
        """
        Get summary timeline
        Returns:
            json, the content of timeline data.
        """
        task_list = []
        timeline_data = []
        with ThreadPoolExecutor() as pool:
            for rank_id, (job_dir, difference_ts) in sub_dirs.items():

                # get step trace
                step_trace_file_name = fr'{job_dir}/step_trace_*.json'
                file_list = get_newest_file(glob.glob(step_trace_file_name))

                if not file_list:
                    logger.warning('Could not find step trace file in %s', job_dir)
                else:
                    task_list.append(pool.submit(self._parse_step_trace_data, file_list,
                                                 rank_id, difference_ts, None,
                                                 merge_model))

                # get overlap analysis
                overlap_file_name = fr'{job_dir}/msprof_*.json'
                file_list = get_newest_file(glob.glob(overlap_file_name))

                if not file_list:
                    logger.warning('Could not find overlap analysis file in %s', job_dir)
                else:
                    task_list.append(pool.submit(self._parse_overlap_analysis_data, file_list,
                                                 rank_id, difference_ts))

            all_done = list(range(len(task_list)))
            while all_done:
                for ind, t in enumerate(task_list):
                    if ind in all_done and t.done():
                        timeline_data.extend(t.result())
                        all_done.remove(ind)

        return timeline_data

    def _get_detail_timeline_data(self, sub_dirs, model_list, merge_model, scope_name):
        """
        Get detail timeline
        Returns:
            json, the content of timeline data.
        """

        timeline_data = []
        task_list = []

        _, model_merged = self._get_models(sub_dirs)
        model_list_all = list(model_merged)
        if model_list_all:
            model_list_all.sort()
        if model_list:
            model_list.sort()
        if model_list_all == model_list:
            model_list = None

        with ThreadPoolExecutor() as pool:
            for rank_id, (job_dir, difference_ts) in sub_dirs.items():
                all_scope_data = []  # 所有带scope的算子

                # get step_trace data
                step_trace_file_name = fr'{job_dir}/step_trace_*.json'
                file_list_step_trace = get_newest_file(glob.glob(step_trace_file_name))
                if not file_list_step_trace:
                    logger.warning('Could not find step trace file in %s', job_dir)
                else:
                    task_list.append(pool.submit(self._parse_step_trace_data, file_list_step_trace,
                                                 rank_id, difference_ts, model_list, merge_model))

                # get Ascend Hardware 、Hccl、CANN、overlap
                msprof_file_name = fr'{job_dir}/msprof_*.json'
                file_list_msprof = get_newest_file(glob.glob(msprof_file_name))
                if not file_list_msprof:
                    logger.warning('Could not find msprof file in %s', job_dir)
                else:
                    ascend_timeline, scope_data = self._parse_msprof_data(file_list_msprof,
                                                                          rank_id, difference_ts, model_list,
                                                                          scope_name)
                    timeline_data.extend(ascend_timeline)
                    all_scope_data.extend(scope_data)

                if not model_list:
                    # get cpu op
                    cpu_op_file_name = fr'{self._profiling_dir}/cpu_op_execute_timestamp_{rank_id}.txt'
                    file_list = glob.glob(cpu_op_file_name)

                    if not file_list:
                        logger.warning('Could not find cpu op file in %s', job_dir)
                    else:
                        cpu_timeline, scope_data = self.parse_cpu_timeline(get_newest_file(file_list),
                                                                           rank_id, difference_ts, scope_name)
                        timeline_data.extend(cpu_timeline)
                        all_scope_data.extend(scope_data)

                # parse scope info
                task_list.append(pool.submit(self._parse_scope_info, all_scope_data,
                                             rank_id, difference_ts))

            all_done = list(range(len(task_list)))
            while all_done:
                for ind, t in enumerate(task_list):
                    if ind in all_done and t.done():
                        timeline_data.extend(t.result())
                        all_done.remove(ind)

        return timeline_data

    def _get_models(self, sub_dirs):
        """
        Get all models
        """
        model_dict = {}
        model_merged = set()
        for rank_id, (job_dir, _) in sub_dirs.items():
            step_trace_file_name = fr'{job_dir}/step_trace_*.csv'
            file_list = get_newest_file(glob.glob(step_trace_file_name))
            if not file_list:
                continue
            model_set = set()
            with open(file_list[0], 'r', newline='') as fr:
                reader = csv.DictReader(fr, delimiter=',', quotechar='"')
                for row in reader:
                    model_id = row.get('Model ID')
                    if model_id:
                        model_set.add(float(model_id))

            model_dict[rank_id] = model_set
            model_merged.update(model_set)
        return model_dict, model_merged



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The minddata pipeline proposer."""
import os
from collections import OrderedDict

from mindinsight.profiler.common.exceptions.exceptions import \
    ProfilerParamValueErrorException
from mindinsight.profiler.common.util import get_options
from mindinsight.profiler.proposer.allproposers.base_proposer import Proposer


class MinddataPipelineProposer(Proposer):
    """
    The minddata pipeline proposer.

    Args:
        profiling_dir (str): The directory in which the parsed profiling data
            resides.
        device_id (str): The device ID.
    """
    _general_label = "minddata_pipeline-general"
    _map_op_label = 'minddata_pipeline-map_op'
    _generator_op_label = 'minddata_pipeline-generator_op'
    _dataset_op_label = 'minddata_pipeline-dataset_op'
    _batch_op_label = 'minddata_pipeline-batch_op'

    def __init__(self, profiling_dir, rank_id):
        super().__init__(profiling_dir, rank_id)
        self.__proposer_type = "minddata_pipeline"
        self.__proposal_dict = OrderedDict()

    def analyze(self, options=None):
        """
        Analyse and get proposal.

        Args:
            options (dict): The options for analysis. Default: None.

        Returns:
            OrderedDict, the proposal for minddata pipeline.
        """
        threshold = self._get_threshold(options)
        condition = {
            'filter_condition': {
                'threshold': threshold
            }
        }
        if os.path.exists(os.path.join(self.profiling_path, "minddata_pipeline_raw_{}.csv".format(self.rank_id))):
            analyser_result = self.get_analyser_result(
                self.__proposer_type, condition=condition
            )
            self._organize_proposal_result(analyser_result)

        return self.__proposal_dict if self.__proposal_dict else None

    def _get_threshold(self, options):
        """
        Get the threshold of the minddata pipeline queue usage rate.

        Args:
            options (dict): The options for analysis.

        Returns:
            list[float], the threshold of the minddata pipeline queue usage rate.

        Raises:
            ProfilerParamValueErrorException: If the threshold is invalid.
        """
        options = get_options(options)
        pipeline_options = options.get(self.__proposer_type)
        threshold = None
        if pipeline_options:
            threshold = options.get('threshold')
        if threshold is None:
            threshold = [0.8, 0.6]
        if not isinstance(threshold, list) or len(threshold) != 2:
            raise ProfilerParamValueErrorException('The threshold is invalid.')
        return threshold

    def _organize_proposal_result(self, analyser_result):
        """
        Organize the proposal result.

        Args:
            analyser_result (dict): The result data of the minddata pipeline
                analyser.
        """
        all_op_names = []
        dataset_op_names = []
        generator_op_names = []
        batch_op_names = []
        map_op_names = []
        pipeline_op_infos = analyser_result.get('object')
        for op_info in pipeline_op_infos:
            op_id = op_info[0]
            op_type = op_info[1]
            op_name = '_'.join([op_type, str(op_id)])
            all_op_names.append(op_name)

            children_ids = op_info[-1]
            if not children_ids:
                dataset_op_names.append(op_name)
            elif op_type == 'MapOp':
                map_op_names.append(op_name)
            elif op_type == 'GeneratorOp':
                generator_op_names.append(op_name)
            elif op_type == 'BatchOp':
                batch_op_names.append(op_name)

        # According to the execution order of data processing operators in the script, the op_names should be reversed.
        if all_op_names:
            self.__proposal_dict[self._general_label] = ['/'.join(all_op_names[::-1])]
        if dataset_op_names:
            self.__proposal_dict[self._dataset_op_label] = [
                '/'.join(dataset_op_names[::-1])
            ]
        if generator_op_names:
            self.__proposal_dict[self._generator_op_label] = [
                '/'.join(generator_op_names[::-1])
            ]
        if map_op_names:
            self.__proposal_dict[self._map_op_label] = ['/'.join(map_op_names[::-1])]
        if batch_op_names:
            self.__proposal_dict[self._batch_op_label] = [
                '/'.join(batch_op_names[::-1])
            ]
        if self.__proposal_dict:
            parent_id = pipeline_op_infos[0][-2]
            if parent_id is None:
                self.__proposal_dict = OrderedDict()



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Debugger server factory."""
import threading

from mindinsight.debugger.common.utils import DebuggerServerMode
from mindinsight.debugger.debugger_services.debugger_offline_server import DebuggerOfflineServer
from mindinsight.debugger.debugger_services.debugger_online_server import DebuggerOnlineServer


class DebuggerServerFactory:
    """Create debugger server according to debugger mode."""

    _lock = threading.Lock()
    _instance = None

    def __init__(self):
        self._server_map = {
            DebuggerServerMode.ONLINE.value: DebuggerOnlineServer,
            DebuggerServerMode.OFFLINE.value: DebuggerOfflineServer
        }

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def get_debugger_server(self, cache_store, context):
        """
        Get debugger server according to debugger_context and cache_store.

        Args:
            cache_store (DebuggerCacheStore): Cache store for debugger server.
            context (DebuggerServerContext): Context for initialize debugger server.

        Returns:
            DebuggerServerBase, Debugger server object.
        """
        dbg_server = None
        dbg_server_class = self._server_map.get(context.dbg_mode)
        if dbg_server_class:
            dbg_server = dbg_server_class(cache_store, context)
        return dbg_server


class DebuggerServerContext:
    """
    Debugger server context.

    Args:
        dbg_mode (str): The debugger mode. Optional: `online` or `offline`.
        train_job (str): The relative directory of debugger dump data for one training.
            Used only when dbg_mode is `offline.`
        dbg_dir (str): The base directory of debugger dump data for one training.
            Used only when dbg_mode is `offline.`
        hostname (str): The hostname used for online debugger server.
            Used only when dbg_mode is `online.`
    """
    def __init__(self, dbg_mode, train_job=None, dbg_dir=None, hostname=None):
        self._dbg_mode = dbg_mode
        self._train_job = train_job
        self._dbg_dir = dbg_dir
        self.hostname = hostname

    @property
    def dbg_mode(self):
        """Property of debugger mode."""
        return self._dbg_mode

    @property
    def dbg_dir(self):
        """Property of debugger mode."""
        return self._dbg_dir

    @property
    def train_job(self):
        """The property of train job."""
        return self._train_job



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Image container."""
from mindinsight.datavisual.proto_files.mindinsight_summary_pb2 import Summary


class ImageContainer:
    """
    Container for image to allow pickling.

    Args:
        image_message (Summary.Image): Image proto buffer message.
    """
    def __init__(self, image_message: Summary.Image):
        self.height = image_message.height
        self.width = image_message.width
        self.colorspace = image_message.colorspace
        self.encoded_image = image_message.encoded_image



# Copyright 2020-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Predefined watchpoints.

This module predefine recommend watchpoints.
"""
import math
import queue as Queue

from mindinsight.debugger.conditionmgr.conditionmgr import ConditionMgr
from mindinsight.debugger.conditionmgr.condition import TargetTypeEnum
from mindinsight.debugger.conditionmgr.condition import ConditionIdEnum
from mindinsight.debugger.conditionmgr.condition import ActivationFuncEnum
from mindinsight.debugger.conditionmgr.common.utils import NodeBasicInfo
from mindinsight.debugger.common.log import LOGGER as logger
from mindinsight.conf import settings
from mindinsight.debugger.stream_cache.node_type_identifier import is_gradient_node
from mindinsight.debugger.stream_cache.watchpoint import WatchNodeTree


class _WatchPointData:
    """
    WatchPoint data container

    Args:
        watch_condition (dict): The dict of watch conditions.
        watch_nodes (list[NodeBasicInfo]): The list of node basic info.
        name (str): The name of watchpoint.
    """
    def __init__(self, watch_condition, watch_nodes, name):
        self.watch_condition = watch_condition
        self.watch_nodes = watch_nodes
        self.name = name

    def get_watch_condition_dict(self):
        return {
            "id": self.watch_condition.get("condition"),
            "params": [{
                "name": param.get_parameter_name(),
                "value": param.value
            } for param in self.watch_condition.get("params")]
        }


class _ConditionParameterValue:
    """Condition parameter data container"""
    def __init__(self, parameter, value):
        self.parameter = parameter
        self.value = value

    def get_parameter_name(self):
        return self.parameter.name


def recommend_watchpoints(condition_mgr: ConditionMgr, multi_card_graph_stream, condition_context):
    """
    Recommend watchpoints.

    Args:
        condition_mgr (ConditionMgr): Condition manager instance.
        multi_card_graph_stream (GraphHandler): Multi card graph handler instance.
        condition_context (ConditionContext): Context for condition.

    Returns:
        list[WatchPointData], watch points to be created.
    """
    watch_points = []

    if not multi_card_graph_stream.has_graph():
        logger.warning("Given graph is None.")
        return watch_points

    if not settings.ENABLE_RECOMMENDED_WATCHPOINTS:
        return watch_points

    # add weight watch points
    merged_info = get_basic_node_info(TargetTypeEnum.WEIGHT.value, multi_card_graph_stream)
    _recommend_weight_initialization(merged_info, condition_mgr, watch_points, condition_context)
    _recommend_weight_change_too_large(merged_info, condition_mgr, watch_points, condition_context)

    # Because we cannot identify trainable weights currently, weight_no_change and weight_change_too_small will not be
    # recommended.
    trainable_weight_nodes = []
    _recommend_weight_not_changed(condition_mgr, trainable_weight_nodes, watch_points, condition_context)
    _recommend_weight_change_too_small(condition_mgr, trainable_weight_nodes, watch_points, condition_context)

    # add gradient watch points
    merged_info = get_basic_node_info(TargetTypeEnum.GRADIENT.value, multi_card_graph_stream)
    _recommend_gradient_vanishing(merged_info, condition_mgr, watch_points, condition_context)

    # add tensor watch points
    merged_info = get_basic_node_info(TargetTypeEnum.TENSOR.value, multi_card_graph_stream)
    _recommend_operator_overflow(merged_info, condition_mgr, watch_points, condition_context)
    _recommend_tensor_overflow(merged_info, condition_mgr, watch_points, condition_context)
    _recommend_tensor_all_zero(merged_info, condition_mgr, watch_points, condition_context)

    # add activation watch points
    merged_info = get_basic_node_info(TargetTypeEnum.ACTIVATION.value, multi_card_graph_stream,
                                      ActivationFuncEnum.TANH.value)
    _recommend_activation_range(merged_info, condition_mgr, watch_points, condition_context,
                                ActivationFuncEnum.TANH.value)

    merged_info = get_basic_node_info(TargetTypeEnum.ACTIVATION.value, multi_card_graph_stream,
                                      ActivationFuncEnum.SIGMOID.value)
    _recommend_activation_range(merged_info, condition_mgr, watch_points, condition_context,
                                ActivationFuncEnum.SIGMOID.value)

    merged_info = get_basic_node_info(TargetTypeEnum.ACTIVATION.value, multi_card_graph_stream,
                                      [ActivationFuncEnum.RELU.value, ActivationFuncEnum.RELUV2.value])
    _recommend_activation_range(merged_info, condition_mgr, watch_points, condition_context,
                                ActivationFuncEnum.RELU.value)
    return watch_points


def _recommend_tensor_all_zero(basic_info_nodes, condition_mgr, watch_points, condition_context):
    """Recommend tensor all zero watchpoint."""
    if not basic_info_nodes:
        return
    if not condition_mgr.has_condition(ConditionIdEnum.TENSOR_ALL_ZERO.value, condition_context):
        return
    condition = condition_mgr.get_condition(condition_id=ConditionIdEnum.TENSOR_ALL_ZERO.value)
    tensor_all_zero_watchpoint = _WatchPointData(
        watch_condition={
            "condition": condition.id,
            "params": [_ConditionParameterValue(
                parameter=condition.get_parameter_definition("zero_percentage_ge"),
                value=100  # set default value to 100
            )]
        },
        watch_nodes=basic_info_nodes.copy(),
        name='recommend_tensor_all_zero_watchpoint'
    )
    watch_points.append(tensor_all_zero_watchpoint)


def _recommend_tensor_overflow(basic_info_nodes, condition_mgr, watch_points, condition_context):
    """Recommend tensor general overflow watchpoint."""
    if not basic_info_nodes:
        return
    if not condition_mgr.has_condition(ConditionIdEnum.TENSOR_OVERFLOW.value, condition_context):
        return

    condition = condition_mgr.get_condition(condition_id=ConditionIdEnum.TENSOR_OVERFLOW.value)
    overflow_watchpoint = _WatchPointData(
        watch_condition={
            "condition": condition.id,
            "params": []
        },
        watch_nodes=basic_info_nodes.copy(),
        name='recommend_tensor_overflow_watchpoint'
    )
    watch_points.append(overflow_watchpoint)


def _recommend_operator_overflow(basic_info_nodes, condition_mgr, watch_points, condition_context):
    """Recommend tensor overflow watchpoint."""
    if not basic_info_nodes:
        return
    if not condition_mgr.has_condition(ConditionIdEnum.OPERATOR_OVERFLOW.value, condition_context):
        return

    condition = condition_mgr.get_condition(condition_id=ConditionIdEnum.OPERATOR_OVERFLOW.value)
    overflow_d_watchpoint = _WatchPointData(
        watch_condition={
            "condition": condition.id,
            "params": []
        },
        watch_nodes=basic_info_nodes.copy(),
        name='recommend_operator_overflow_watchpoint'
    )
    watch_points.append(overflow_d_watchpoint)


def _recommend_gradient_vanishing(basic_info_nodes, condition_mgr, watch_points, condition_context):
    """Recommend gradient vanishing watchpoint."""
    if not basic_info_nodes:
        return
    if not condition_mgr.has_condition(ConditionIdEnum.GRADIENT_VANISHING.value, condition_context):
        return

    condition = condition_mgr.get_condition(condition_id=ConditionIdEnum.GRADIENT_VANISHING.value)
    gradient_vanishing_watchpoint = _WatchPointData(
        watch_condition={
            "condition": condition.id,
            "params": [_ConditionParameterValue(
                parameter=condition.get_parameter_definition("abs_mean_lt"),
                value=1e-9  # set default value to 1e-9
            )]
        },
        watch_nodes=basic_info_nodes.copy(),
        name='recommend_gradient_vanishing_watchpoint'
    )
    watch_points.append(gradient_vanishing_watchpoint)


def _recommend_weight_change_too_small(condition_mgr, trainable_weight_nodes, watch_points, condition_context):
    """Recommend weight change too small watchpoint."""
    if not trainable_weight_nodes:
        return
    if not condition_mgr.has_condition(ConditionIdEnum.WEIGHT_CHANGE_TOO_SMALL.value, condition_context):
        return

    condition = condition_mgr.get_condition(condition_id=ConditionIdEnum.WEIGHT_CHANGE_TOO_SMALL.value)
    weight_change_too_small_watchpoint = _WatchPointData(
        watch_condition={
            "condition": condition.id,
            "params": [
                _ConditionParameterValue(
                    parameter=condition.get_parameter_definition("abs_mean_update_ratio_lt"),
                    value=1.0e-4  # set default value to 1.0e-4
                ),
            ]
        },
        watch_nodes=trainable_weight_nodes,
        name='recommend_weight_change_too_small_watchpoint'
    )
    watch_points.append(weight_change_too_small_watchpoint)


def _recommend_weight_not_changed(condition_mgr, trainable_weight_nodes, watch_points, condition_context):
    """Recommend weight not changed watchpoint."""
    if not trainable_weight_nodes:
        return
    if not condition_mgr.has_condition(ConditionIdEnum.WEIGHT_NOT_CHANGED.value, condition_context):
        return

    condition = condition_mgr.get_condition(condition_id=ConditionIdEnum.WEIGHT_NOT_CHANGED.value)
    weight_no_change_watchpoint = _WatchPointData(
        watch_condition={
            "condition": condition.id,
            "params": [
                _ConditionParameterValue(
                    parameter=condition.get_parameter_definition("rtol"),
                    value=1.0e-5  # set default value to 1.0e-5
                ),
                _ConditionParameterValue(
                    parameter=condition.get_parameter_definition("atol"),
                    value=1.0e-8  # set default value to 1.0e-8
                ),
            ]
        },
        watch_nodes=trainable_weight_nodes,
        name='recommend_weight_not_changed_watchpoint'
    )
    watch_points.append(weight_no_change_watchpoint)


def _recommend_weight_change_too_large(basic_info_nodes, condition_mgr, watch_points, condition_context):
    """Recommend weight change too large watchpoint."""
    if not basic_info_nodes:
        return
    if not condition_mgr.has_condition(ConditionIdEnum.WEIGHT_CHANGE_TOO_LARGE.value, condition_context):
        return

    condition = condition_mgr.get_condition(condition_id=ConditionIdEnum.WEIGHT_CHANGE_TOO_LARGE.value)
    weight_initialization_watchpoint = _WatchPointData(
        watch_condition={
            "condition": condition.id,
            "params": [_ConditionParameterValue(
                parameter=condition.get_parameter_definition("abs_mean_update_ratio_gt"),
                value=1  # set default value to 1
            )]
        },
        watch_nodes=basic_info_nodes.copy(),
        name='recommend_weight_change_too_large_watchpoint'
    )
    watch_points.append(weight_initialization_watchpoint)


def _recommend_weight_initialization(basic_info_nodes, condition_mgr, watch_points, condition_context):
    """Recommend weight initialization watchpoint."""
    if not basic_info_nodes:
        return
    if not condition_mgr.has_condition(ConditionIdEnum.WEIGHT_INITIALIZATION.value, condition_context):
        return

    condition = condition_mgr.get_condition(condition_id=ConditionIdEnum.WEIGHT_INITIALIZATION.value)
    weight_initialization_watchpoint = _WatchPointData(
        watch_condition={
            "condition": condition.id,
            "params": [_ConditionParameterValue(
                parameter=condition.get_parameter_definition("zero_percentage_ge"),
                value=100  # set default value to 100
            )]
        },
        watch_nodes=basic_info_nodes.copy(),
        name='recommend_weight_initialization_watchpoint'
    )
    watch_points.append(weight_initialization_watchpoint)


def _recommend_activation_range(basic_info_nodes, condition_mgr, watch_points, condition_context, activation_func):
    """Recommend activation range watchpoint."""
    if not basic_info_nodes:
        return
    if not condition_mgr.has_condition(ConditionIdEnum.ACTIVATION_RANGE.value, condition_context):
        return
    condition = condition_mgr.get_condition(condition_id=ConditionIdEnum.ACTIVATION_RANGE.value)
    params = _get_recommend_activation_params(condition, activation_func)
    activation_range_watchpoint = _WatchPointData(
        watch_condition={
            "condition": condition.id,
            "params": params
        },
        watch_nodes=basic_info_nodes.copy(),
        name='recommend_{}_activation_range_watchpoint'.format(activation_func.lower())
    )
    watch_points.append(activation_range_watchpoint)


def get_basic_node_info(node_category, multi_card_graph_stream, activation_func=None):
    """Get node merged info."""
    if node_category in [TargetTypeEnum.TENSOR.value, TargetTypeEnum.GRADIENT.value]:
        return _get_top_nodes(node_category, multi_card_graph_stream)
    nodes_for_devices = {}
    has_node = False
    for rank_id, graph_stream in multi_card_graph_stream.graph_handlers.items():
        basic_info_nodes = _get_basic_node_info_by_node_category(node_category, graph_stream, activation_func)
        merged_info = _merge_nodes(basic_info_nodes, graph_stream.whole_graph)
        merged_info = _add_graph_name(merged_info, graph_stream)
        nodes_for_devices[rank_id] = merged_info
        has_node = has_node or merged_info

    if has_node:
        return nodes_for_devices

    return {}


def _get_top_nodes(node_category, multi_card_graph_stream):
    """Get top scope nodes, only for tensor and gradient nodes."""
    nodes_for_devices = {}
    for rank_id, graph_stream in multi_card_graph_stream.graph_handlers.items():
        nodes_info = []
        nodes = []
        if node_category == TargetTypeEnum.TENSOR.value:
            graph = graph_stream.get()
            if graph is not None:
                nodes = graph.get('graph', {}).get('nodes', [])
            for node in nodes:
                node_basic_info = NodeBasicInfo(name=node.get('name', ''), full_name=node.get('full_name', ''),
                                                type=node.get('type', ''))
                nodes_info.append(node_basic_info)
        else:
            pattern = {'name': TargetTypeEnum.GRADIENT.value}
            nodes = graph_stream.search_in_graph(pattern)
            for node in nodes:
                if not is_gradient_node(node):
                    continue
                node_basic_info = NodeBasicInfo(name=node.name, full_name=node.full_name, type=node.type)
                nodes_info.append(node_basic_info)
        if nodes_for_devices.get(rank_id) is None:
            nodes_for_devices[rank_id] = []
        nodes_for_devices[rank_id].extend(_add_graph_name(nodes_info, graph_stream))
    return nodes_for_devices


def _get_basic_node_info_by_node_category(node_category, graph_stream, activation_func=None):
    """Get node basic info by node category."""
    pattern = {'node_category': node_category}
    if activation_func:
        pattern['condition'] = {'activation_func': activation_func}
    all_graph_nodes = graph_stream.search_in_graph(pattern)
    return all_graph_nodes


def _convert_tree_to_node_list(node_tree, node_list):
    """Convert WatchNodeTree to Node list."""
    if node_tree.watch_status in [WatchNodeTree.NOT_WATCH, WatchNodeTree.INVALID]:
        logger.debug("The watch_status of node: %s is not_watch or invalid.", node_tree.node_name)
        return
    if node_tree.watch_status == WatchNodeTree.TOTAL_WATCH:
        node_basic_info = NodeBasicInfo(name=node_tree.node_name, full_name=node_tree.full_name,
                                        type=node_tree.node_type)
        node_list.append(node_basic_info)
        return
    if node_tree.watch_status == WatchNodeTree.PARTIAL_WATCH:
        for _, sub_tree in node_tree.get_children():
            _convert_tree_to_node_list(sub_tree, node_list)


def _update_watch_status(node_tree, graph):
    """Update the watch_status, if all sub_nodes of a WatchNodeTree are total_watch,
    then the WatchNodeTree is changed to total_watch status."""
    tmp_node_queue = Queue.Queue()
    tmp_node_queue.put(node_tree)

    # watch node list in layer order
    watch_tree_list = []
    while not tmp_node_queue.empty():
        cur_tree = tmp_node_queue.get()
        watch_tree_list.append(cur_tree)
        for _, sub_tree in cur_tree.get_children():
            tmp_node_queue.put(sub_tree)

    # update the watch_status from bottom to top
    while watch_tree_list:
        cur_tree = watch_tree_list.pop()
        node_name = cur_tree.node_name
        logger.debug("Update status of node: %s.", node_name)

        # if node_name is "", it is the root node, which is not in normal_node_map
        if not node_name:
            continue
        sub_count = graph.normal_node_map.get(node_name).subnode_count

        # if the children_count of WatchNodeTree is less than the responding subnode_count in the graph,
        # its watch_status must be partial_watch
        if cur_tree.get_children_count() < sub_count:
            continue
        is_all_chosen = True
        for _, sub_tree in cur_tree.get_children():
            if sub_tree.watch_status != WatchNodeTree.TOTAL_WATCH:
                is_all_chosen = False
                break

        if is_all_chosen:
            cur_tree.watch_status = WatchNodeTree.TOTAL_WATCH


def _merge_nodes(leaf_nodes, graph):
    """Merge nodes in one graph."""
    watch_node_tree = WatchNodeTree()
    for node in leaf_nodes:
        watch_node_tree.add_node(node.name, node.type, node.full_name)
    _update_watch_status(watch_node_tree, graph)
    out_nodes = []
    _convert_tree_to_node_list(watch_node_tree, out_nodes)
    logger.debug("out_nodes: %s", out_nodes)
    return out_nodes


def _add_graph_name(nodes, graph_stream):
    """Add graph_name in node.name."""
    if len(graph_stream.graph_names) > 1:
        return nodes
    graph_name = graph_stream.graph_names[0]
    output_nodes = []
    for node in nodes:
        node_basic_info = graph_stream.construct_node_basic_info(
            full_name=node.full_name, graph_name=graph_name, node_name=node.name, node_type=node.type)
        output_nodes.append(node_basic_info)
    return output_nodes


def _sigmoid(value):
    """Calculate the sigmoid of value."""
    return 1.0 / (1.0 + math.exp(-value))


def _get_recommend_activation_params(condition, activation_func):
    """Get recommend params for tanh, sigmoid and relu activation function."""
    params = []
    if activation_func == ActivationFuncEnum.TANH.value:
        # The recommend params for Tanh: The percentage of value in range (tanh(-8.8), tanh(8.8)) is lower than 0.1%
        params = [
            _ConditionParameterValue(
                parameter=condition.get_parameter_definition("range_percentage_lt"),
                value=0.1
            ),
            _ConditionParameterValue(
                parameter=condition.get_parameter_definition("range_start_inclusive"),
                value=math.tanh(-8.8)
            ),
            _ConditionParameterValue(
                parameter=condition.get_parameter_definition("range_end_inclusive"),
                value=math.tanh(8.8)
            )]
    if activation_func == ActivationFuncEnum.SIGMOID.value:
        # The recommend params for Sigmoid:
        # The percentage of value in range (sigmoid(-16.2)), sigmoid(16.2)) is lower than 0.1%
        params = [
            _ConditionParameterValue(
                parameter=condition.get_parameter_definition("range_percentage_lt"),
                value=0.1
            ),
            _ConditionParameterValue(
                parameter=condition.get_parameter_definition("range_start_inclusive"),
                value=_sigmoid(-16.2)
            ),
            _ConditionParameterValue(
                parameter=condition.get_parameter_definition("range_end_inclusive"),
                value=_sigmoid(16.2)
            )]
    if activation_func == ActivationFuncEnum.RELU.value:
        # The recommend params for ReLU:
        # The percentage of value in range (-1, 0) is greater than 99.9%
        params = [
            _ConditionParameterValue(
                parameter=condition.get_parameter_definition("range_percentage_gt"),
                value=99.9
            ),
            _ConditionParameterValue(
                parameter=condition.get_parameter_definition("range_start_inclusive"),
                value=-1
            ),
            _ConditionParameterValue(
                parameter=condition.get_parameter_definition("range_end_inclusive"),
                value=0
            )]
    return params



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The definition of tensor stream."""
from abc import abstractmethod, ABC
from enum import Enum

import numpy as np

from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import NUMPY_TYPE_MAP, MAX_SINGLE_TENSOR_CACHE_BYTES
from mindinsight.domain.graph.proto.ms_graph_pb2 import DataType
from mindinsight.utils.tensor import TensorUtils

NPY_HEADER_LENGTH = 128


class TensorStatusEnum(Enum):
    """Tensor status."""
    CACHED = "cached"
    UNCACHED = "uncached"
    EMPTY = 'empty'
    OVERSIZE = "oversize"


class DownloadStatusEnum(Enum):
    """Tensor status."""
    PENDING = "pending"
    SENDING = "sending"


class BaseTensor(ABC):
    """Tensor data structure."""

    def __init__(self, step=0):
        self._step = step
        self._stats = None

    @property
    @abstractmethod
    def name(self):
        """The property of tensor name."""

    @property
    @abstractmethod
    def dtype(self):
        """The property of tensor dtype."""

    @property
    @abstractmethod
    def shape(self):
        """The property of tensor shape."""

    @property
    @abstractmethod
    def value(self):
        """The property of tensor value."""

    @property
    @abstractmethod
    def download_size(self):
        """The property of download size."""

    @property
    @abstractmethod
    def status(self):
        """The property of tensor status."""

    @property
    def empty(self):
        """If the tensor value is valid."""
        return self.value is None

    @property
    def stats(self):
        """The property of tensor stats."""
        return self._stats

    @stats.setter
    def stats(self, stats):
        """
        Update tensor stats.

        Args:
            stats (Statistics): Instance of Statistics.
        """
        self._stats = stats

    def get_tensor_serializable_value_by_shape(self, shape=None):
        """
        Get tensor value info by shape.

        Args:
            shape (tuple): The specified range of tensor value.

        Returns:
            dict, the specified tensor value and value statistics.
        """
        tensor_value = self.get_tensor_value_by_shape(shape)
        res = {}
        # the type of tensor_value is one of None, np.ndarray or str
        if isinstance(tensor_value, np.ndarray):
            res['value'] = tensor_value.tolist()
        else:
            res['value'] = tensor_value
        res['statistics'] = self.get_tensor_statistics()
        return res

    @abstractmethod
    def get_tensor_value_by_shape(self, shape=None):
        """Abstract method."""

    @abstractmethod
    def get_tensor_statistics(self):
        """Abstract method."""

    def _to_dict(self):
        """Get tensor info in dict format."""
        res = {
            'full_name': self.name,
            'step': self._step,
            'dtype': self.dtype,
            'shape': self.shape,
            'bytes': self.download_size,
            'tensor_status': self.status,
            'has_prev_step': False
        }
        return res

    def get_basic_info(self):
        """Return basic info about tensor info."""
        if self.status == TensorStatusEnum.EMPTY.value:
            value = None
        else:
            value = 'click to view'
        res = self._to_dict()
        res['value'] = value
        return res

    def get_full_info(self, shape=None):
        """Get tensor info with value."""
        res = self._to_dict()
        value_info = self.get_tensor_serializable_value_by_shape(shape)
        res.update(value_info)
        return res


class OpTensor(BaseTensor):
    """
    Tensor data structure for operator Node.

    Args:
        tensor_proto (TensorProto): Tensor proto contains tensor basic info.
        tensor_content (byte): Tensor content value in byte format.
        step (int): The step of the tensor.
    """
    max_number_data_show_on_ui = 100000

    def __init__(self, name, tensor_base=None, tensor_stats=None, tensor_content=None, step=0):
        # the type of tensor_proto is TensorProto
        super(OpTensor, self).__init__(step)
        self._name = name
        self._tensor_base = tensor_base
        self._shape = tensor_base.get('shape') if tensor_base else []
        self._value = self.to_numpy(tensor_content)
        self._download_size = tensor_base.get('data_size') + NPY_HEADER_LENGTH if tensor_base else 0
        self._stats = tensor_stats
        self._tensor_comparison = None
        self._status = self.get_status()

    @property
    def name(self):
        """The property of tensor name."""
        return self._name

    @property
    def tensor_base(self):
        """The property of tensor base."""
        return self._tensor_base

    @property
    def step(self):
        """The step of the tensor value."""
        return self._step

    @property
    def dtype(self):
        """The property of tensor dtype."""
        return DataType.Name(self._tensor_base.get('dtype') if self._tensor_base else 0)

    @property
    def shape(self):
        """The property of tensor shape."""
        dims = list(self._shape)
        if dims == [0]:
            dims = []
        return dims

    @property
    def value(self):
        """The property of tensor value."""
        return self._value

    @property
    def nbytes(self):
        """The property of tensor value."""
        return self._value.nbytes if self._value is not None else 0

    @property
    def download_size(self):
        """The property of download size."""
        return self._download_size

    @property
    def status(self):
        """The property of tensor status."""
        return self._status

    @property
    def tensor_comparison(self):
        """The property of tensor_comparison."""
        return self._tensor_comparison

    def get_status(self):
        """Set the tensor status."""
        res = TensorStatusEnum.EMPTY.value if self.tensor_base is None or self.tensor_base.get(
            'data_size') == 0 else TensorStatusEnum.OVERSIZE.value if self.tensor_base.get(
                'data_size') >= MAX_SINGLE_TENSOR_CACHE_BYTES else TensorStatusEnum.UNCACHED.value
        return res if self.value is None else TensorStatusEnum.CACHED.value

    def calculate_stats(self):
        """Calculate the tensor statistics."""
        self._stats = TensorUtils.get_statistics_from_tensor(self._value)

    def clean_tensor_value(self, oversize=True, remain_scalar=False):
        """Clean the tensor."""
        if oversize:
            self._status = TensorStatusEnum.OVERSIZE.value
        if remain_scalar and not self.shape and self._status != TensorStatusEnum.EMPTY.value:
            return
        self._value = None
        if self._status == TensorStatusEnum.CACHED.value:
            self._status = TensorStatusEnum.UNCACHED.value

    def to_numpy(self, tensor_content):
        """
        Construct tensor content from byte to numpy.

        Args:
            tensor_content (byte): The tensor content.

        Returns:
            Union[None, np.ndarray], the value of the tensor.
        """
        tensor_value = None
        if tensor_content:
            np_type = NUMPY_TYPE_MAP.get(self.dtype)
            tensor_value = np.frombuffer(tensor_content, dtype=np_type)
            tensor_value = tensor_value.reshape(self.shape)
        return tensor_value

    def get_tensor_statistics(self):
        """
        Get Tensor statistics.

        Returns:
            dict, overall statistics.
        """
        if self.stats:
            return TensorUtils.get_overall_statistic_dict(self.stats)
        if self.empty:
            return {}
        self.stats = TensorUtils.get_statistics_from_tensor(self.value)
        return TensorUtils.get_overall_statistic_dict(self.stats)

    def update_tensor_comparisons(self, tensor_comparison):
        """
        Update tensor comparison for tensor.

        Args:
            tensor_comparison (TensorComparison) instance of TensorComparison.

        """
        self._tensor_comparison = tensor_comparison

    def get_tensor_value_by_shape(self, shape=None):
        """
        Get tensor value by shape.

        Args:
            shape (tuple): The specified shape.

        Returns:
            Union[None, str, numpy.ndarray], the value of parsed tensor.
        """
        if self._status == TensorStatusEnum.OVERSIZE.value:
            log.info("%s is too large to show on UI.", self.name)
            return "Too large to show on UI."
        if self._value is None:
            log.warning("%s has no value yet.", self.name)
            return None
        if shape is None or not isinstance(shape, tuple):
            log.info("Get the whole tensor value with shape is %s", shape)
            return self._value
        if len(shape) != len(self.shape):
            log.error("Invalid shape. Received: %s, tensor shape: %s", shape, self.shape)
            raise DebuggerParamValueError("Invalid shape. Shape unmatched.")
        try:
            value = self._value[shape]
        except IndexError as err:
            log.error("Invalid shape. Received: %s, tensor shape: %s", shape, self.shape)
            log.exception(err)
            raise DebuggerParamValueError("Invalid shape. Shape unmatched.")
        if isinstance(value, np.ndarray):
            if value.size > self.max_number_data_show_on_ui:
                log.info("The tensor size is %d, which is too large to show on UI.", value.size)
                value = "Too large to show."
        else:
            value = np.asarray(value)
        return value


class ConstTensor(BaseTensor):
    """Tensor data structure for Const Node."""
    _STRING_TYPE = 'DT_STRING'
    _DT_TYPE = 'DT_TYPE'

    def __init__(self, const_proto):
        # the type of const_proto is NamedValueProto
        super(ConstTensor, self).__init__()
        self._const_proto = const_proto
        self._value = self.generate_value_from_proto(const_proto.value)
        self._status = TensorStatusEnum.CACHED.value if self._value is not None else TensorStatusEnum.EMPTY.value
        self._stats = None
        self._tensor_base = {'dtype': DataType.Name(self._const_proto.value.dtype),
                             'shape': [],
                             'data_size': self.nbytes}

    def set_step(self, step):
        """Set step value."""
        self._step = step

    @property
    def name(self):
        """The property of tensor name."""
        return self._const_proto.key + ':0'

    @property
    def full_name(self):
        """The property of tensor full name."""
        return self._const_proto.full_name + ':0'

    @property
    def dtype(self):
        """The property of tensor dtype."""
        return DataType.Name(self._const_proto.value.dtype)

    @property
    def shape(self):
        """The property of tensor shape."""
        return []

    @property
    def value(self):
        """The property of tensor shape."""
        return self._value

    @property
    def status(self):
        """The property of tensor status."""
        return self._status

    @property
    def download_size(self):
        """The property of download size."""
        return self.nbytes

    @property
    def tensor_base(self):
        """The property of tensor_base."""
        return self._tensor_base

    @property
    def nbytes(self):
        """The property of tensor size."""
        return self._value.nbytes if self._value is not None else 0

    def generate_value_from_proto(self, tensor_proto):
        """
        Generate tensor value from proto.

        Args:
            tensor_proto (TensorProto): The tensor proto.

        Returns:
            Union[None, str, np.ndarray], the value of the tensor.
        """
        fields = tensor_proto.ListFields()
        if len(fields) != 2:
            log.warning("Unexpected const proto <%s>.\n Please check offline.", tensor_proto)
        tensor_value = None
        for field_obj, field_value in fields:
            if field_obj.name != 'dtype':
                if tensor_proto.dtype == DataType.DT_TUPLE:
                    tensor_values = []
                    for field_value_element in field_value:
                        value_element = self.generate_value_from_proto(field_value_element)
                        tensor_values.append(value_element)
                    tensor_value = tensor_values
                elif tensor_proto.dtype == DataType.DT_TYPE:
                    tensor_value = DataType.Name(field_value.data_type)
                else:
                    tensor_value = field_value
                break
        if tensor_value is not None and tensor_proto.dtype != self._STRING_TYPE:
            tensor_value = np.array(tensor_value, dtype=NUMPY_TYPE_MAP.get(tensor_proto.dtype))
        return tensor_value

    def get_tensor_value_by_shape(self, shape=None):
        """
        Get tensor value by shape.

        Args:
            shape (tuple): The specified shape.

        Returns:
            Union[None, str, int, float], the value of parsed tensor.
        """
        if shape:
            log.warning("Invalid shape for const value.")
        return self._value

    def calculate_stats(self):
        """Calculate the tensor statistics."""
        if not (self.empty or self.dtype == self._STRING_TYPE or self.dtype == self._DT_TYPE):
            self._stats = TensorUtils.get_statistics_from_tensor(self._value)

    def get_tensor_statistics(self):
        """
        Get Tensor statistics.

        Returns:
            dict, overall statistics.
        """
        self.calculate_stats()
        if self.empty or self.dtype == self._STRING_TYPE or self.dtype == self._DT_TYPE:
            log.debug("The tensor dtype is: %s, skip getting statistics.", self.dtype)
            return {}
        statistics = TensorUtils.get_overall_statistic_dict(self._stats)
        return statistics



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to define lineage info model."""
import json
from collections import namedtuple

from google.protobuf.json_format import MessageToDict

from mindinsight.lineagemgr.common.exceptions.exceptions import \
    LineageEventFieldNotExistException, LineageEventNotExistException
from mindinsight.lineagemgr.summary._summary_adapter import organize_graph

Field = namedtuple('Field', ['base_name', 'sub_name'])
FIELD_MAPPING = {
    "summary_dir": Field('summary_dir', None),
    "loss_function": Field("hyper_parameters", 'loss_function'),
    "train_dataset_path": Field('train_dataset', 'train_dataset_path'),
    "train_dataset_count": Field("train_dataset", 'train_dataset_size'),
    "test_dataset_path": Field('valid_dataset', 'valid_dataset_path'),
    "test_dataset_count": Field('valid_dataset', 'valid_dataset_size'),
    "network": Field('algorithm', 'network'),
    "optimizer": Field('hyper_parameters', 'optimizer'),
    "learning_rate": Field('hyper_parameters', 'learning_rate'),
    "epoch": Field('hyper_parameters', 'epoch'),
    "batch_size": Field('hyper_parameters', 'batch_size'),
    "device_num": Field('hyper_parameters', 'device_num'),
    "loss": Field('algorithm', 'loss'),
    "model_size": Field('model', 'size'),
    "dataset_mark": Field('dataset_mark', None)
}


class LineageObj:
    """
    Lineage information class.

    An instance of the class hold lineage information for a training session.

    Args:
        summary_dir (str): Summary log dir.
        kwargs (dict): Params to init the instance.

            - train_lineage (Event): Train lineage object.

            - evaluation_lineage (Event): Evaluation lineage object.

            - dataset_graph (Event): Dataset graph object.

            - user_defined_info (Event): User defined info object.

    Raises:
        LineageEventNotExistException: If train and evaluation event not exist.
        LineageEventFieldNotExistException: If the special event field not exist.
    """
    _name_train_lineage = 'train_lineage'
    _name_evaluation_lineage = 'evaluation_lineage'
    _name_summary_dir = 'summary_dir'
    _name_metric = 'metric'
    _name_hyper_parameters = 'hyper_parameters'
    _name_algorithm = 'algorithm'
    _name_train_dataset = 'train_dataset'
    _name_model = 'model'
    _name_valid_dataset = 'valid_dataset'
    _name_dataset_graph = 'dataset_graph'
    _name_dataset_mark = 'dataset_mark'
    _name_user_defined = 'user_defined'
    _name_model_lineage = 'model_lineage'

    def __init__(self, summary_dir, **kwargs):
        self._lineage_info = {
            self._name_summary_dir: summary_dir
        }
        self._init_lineage()
        self.parse_and_update_lineage(**kwargs)

    def _init_lineage(self):
        """Init lineage info."""
        # train
        self._lineage_info[self._name_model] = {}
        self._lineage_info[self._name_algorithm] = {}
        self._lineage_info[self._name_hyper_parameters] = {}
        self._lineage_info[self._name_train_dataset] = {}

        # eval
        self._lineage_info[self._name_metric] = {}
        self._lineage_info[self._name_valid_dataset] = {}

        # dataset graph
        self._lineage_info[self._name_dataset_graph] = {}

        # user defined
        self._lineage_info[self._name_user_defined] = {}

    def parse_and_update_lineage(self, **kwargs):
        """Parse and update lineage."""
        user_defined_info_list = kwargs.get('user_defined_info', [])
        train_lineage = kwargs.get('train_lineage')
        evaluation_lineage = kwargs.get('evaluation_lineage')
        dataset_graph = kwargs.get('dataset_graph')
        if not any([train_lineage, evaluation_lineage, dataset_graph]):
            raise LineageEventNotExistException()

        # If new train lineage, will clean the lineage saved before.
        if train_lineage is not None or dataset_graph is not None:
            self._init_lineage()

        self._parse_user_defined_info(user_defined_info_list)
        self._parse_train_lineage(train_lineage)
        self._parse_evaluation_lineage(evaluation_lineage)
        self._parse_dataset_graph(dataset_graph)

        self._filtration_result = self._organize_filtration_result()

    @property
    def summary_dir(self):
        """
        Get summary log dir.

        Returns:
            str, the summary log dir.
        """
        return self._lineage_info.get(self._name_summary_dir)

    @property
    def metric(self):
        """
        Get metric information.

        Returns:
            dict, the metric information.
        """
        return self._lineage_info.get(self._name_metric)

    @property
    def user_defined(self):
        """
        Get user defined information.

        Returns:
            dict, the user defined information.
        """
        return self._lineage_info.get(self._name_user_defined)

    @property
    def hyper_parameters(self):
        """
        Get hyperparameters.

        Returns:
            dict, the hyperparameters.
        """
        return self._lineage_info.get(self._name_hyper_parameters)

    @property
    def algorithm(self):
        """
        Get algorithm.

        Returns:
            dict, the algorithm.
        """
        return self._lineage_info.get(self._name_algorithm)

    @property
    def train_dataset(self):
        """
        Get train dataset information.

        Returns:
            dict, the train dataset information.
        """
        return self._lineage_info.get(self._name_train_dataset)

    @property
    def model(self):
        """
        Get model information.

        Returns:
            dict, the model information.
        """
        return self._lineage_info.get(self._name_model)

    @property
    def valid_dataset(self):
        """
        Get valid dataset information.

        Returns:
            dict, the valid dataset information.
        """
        return self._lineage_info.get(self._name_valid_dataset)

    @property
    def dataset_graph(self):
        """
        Get dataset_graph.

        Returns:
            dict, the dataset graph information.
        """
        return self._lineage_info.get(self._name_dataset_graph)

    @property
    def dataset_mark(self):
        """
        Get dataset_mark.

        Returns:
            dict, the dataset mark information.
        """
        return self._lineage_info.get(self._name_dataset_mark)

    @dataset_mark.setter
    def dataset_mark(self, dataset_mark):
        """
        Set dataset mark.

        Args:
            dataset_mark (int): Dataset mark.
        """
        self._lineage_info[self._name_dataset_mark] = dataset_mark
        # update dataset_mark into filtration result
        self._filtration_result[self._name_dataset_mark] = dataset_mark

    def get_summary_info(self, filter_keys: list):
        """
        Get the summary lineage information.

        Returns the content corresponding to the specified field in the filter
        key. The contents of the filter key include `metric`, `hyper_parameters`,
        `algorithm`, `train_dataset`, `valid_dataset` and `model`. You can
        specify multiple filter keys in the `filter_keys`

        Args:
            filter_keys (list): Filter keys.

        Returns:
            dict, the summary lineage information.
        """
        result = {
            self._name_summary_dir: self.summary_dir,
        }

        for key in filter_keys:
            result[key] = getattr(self, key)
        return result

    def to_dataset_lineage_dict(self):
        """
        Returns the dataset part lineage information.

        Returns:
            dict, the dataset lineage information.
        """
        dataset_lineage = {
            key: self._filtration_result.get(key)
            for key in [self._name_summary_dir, self._name_dataset_graph]
        }

        return dataset_lineage

    def to_model_lineage_dict(self):
        """
        Returns the model part lineage information.

        Returns:
            dict, the model lineage information.
        """
        filtration_result = dict(self._filtration_result)
        filtration_result.pop(self._name_dataset_graph)

        model_lineage = dict()
        model_lineage.update({self._name_summary_dir: filtration_result.pop(self._name_summary_dir)})
        model_lineage.update({self._name_model_lineage: filtration_result})

        return model_lineage

    def get_value_by_key(self, key):
        """
        Get the value based on the key in `FIELD_MAPPING` or
            the key prefixed with `metric/` or `user_defined/`.

        Args:
            key (str): The key in `FIELD_MAPPING`
                or prefixed with `metric/` or `user_defined/`.

        Returns:
            object, the value.
        """
        if key.startswith(('metric/', 'user_defined/')):
            key_name, sub_key = key.split('/', 1)
            sub_value_name = self._name_metric if key_name == 'metric' else self._name_user_defined
            sub_value = self._filtration_result.get(sub_value_name)
            if sub_value:
                return sub_value.get(sub_key)
        return self._filtration_result.get(key)

    def _organize_filtration_result(self):
        """
        Organize filtration result.

        Returns:
            dict, the filtration result.
        """
        result = {}
        for key, field in FIELD_MAPPING.items():
            if field.base_name is not None:
                base_attr = getattr(self, field.base_name)
                result[key] = base_attr.get(field.sub_name) \
                    if field.sub_name else base_attr
        # add metric into filtration result
        result[self._name_metric] = self.metric

        result[self._name_user_defined] = self.user_defined
        # add dataset_graph into filtration result
        result[self._name_dataset_graph] = getattr(self, self._name_dataset_graph)

        return result

    def _parse_train_lineage(self, train_lineage):
        """
        Parse train lineage.

        Args:
            train_lineage (Event): Train lineage.
        """
        if train_lineage is None:
            return

        event_dict = MessageToDict(
            train_lineage, preserving_proto_field_name=True
        )
        train_dict = event_dict.get(self._name_train_lineage)
        if train_dict is None:
            raise LineageEventFieldNotExistException(
                self._name_train_lineage
            )

        # when MessageToDict is converted to dict, int64 type is converted
        # to string, so we convert it to an int in python
        if train_dict.get(self._name_model):
            model_size = train_dict.get(self._name_model).get('size')
            if model_size:
                train_dict[self._name_model]['size'] = int(model_size)

        self._lineage_info.update(**train_dict)

    def _parse_evaluation_lineage(self, evaluation_lineage):
        """
        Parse evaluation lineage.

        Args:
            evaluation_lineage (Event): Evaluation lineage.
        """
        if evaluation_lineage is None:
            return

        event_dict = MessageToDict(
            evaluation_lineage, preserving_proto_field_name=True
        )
        evaluation_dict = event_dict.get(self._name_evaluation_lineage)
        if evaluation_dict is None:
            raise LineageEventFieldNotExistException(
                self._name_evaluation_lineage
            )
        self._lineage_info.update(**evaluation_dict)
        metric = self._lineage_info.get(self._name_metric)
        self._lineage_info[self._name_metric] = json.loads(metric) if metric else {}

    def _parse_dataset_graph(self, dataset_graph):
        """
        Parse dataset graph.

        Args:
            dataset_graph (Event): Dataset graph.
        """
        if dataset_graph is not None:
            # convert message to dict
            event_dict = organize_graph(dataset_graph.dataset_graph)
            if event_dict is None:
                raise LineageEventFieldNotExistException(self._name_evaluation_lineage)
            self._lineage_info[self._name_dataset_graph] = event_dict if event_dict else {}

    def _parse_user_defined_info(self, user_defined_info_list):
        """
        Parse user defined info.

        Args:
            user_defined_info_list (list): user defined info list.
        """
        if not user_defined_info_list:
            return
        for user_defined_info in user_defined_info_list:
            self._lineage_info[self._name_user_defined].update(user_defined_info)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Condition manager..

This module provide condition manager function.
"""
from mindinsight.debugger.conditionmgr.condition import Condition
from mindinsight.debugger.conditionmgr.condition import TargetTypeEnum
from mindinsight.debugger.conditionmgr.condition_list import CONDITION_LIST
from mindinsight.debugger.common.log import LOGGER as logger


class ConditionMgr:
    """Condition manager."""

    def __init__(self):
        self.conditions = {}
        self.no_parameter_conditions = []
        self._register_default_conditions()

    def _register_default_conditions(self):
        """Register default condition definitions"""
        self.register_conditions(CONDITION_LIST)

    def register_condition(self, condition):
        """Register conditions into dict"""
        if not condition.parameters:
            self.no_parameter_conditions.append(condition.id)
        self.conditions[condition.id] = condition

    def register_conditions(self, conditions):
        """Register conditions"""
        for condition in conditions:
            self.register_condition(condition)

    def get_condition(self, condition_id) -> Condition:
        """Get condition by condition id"""
        return self.conditions[condition_id]

    def has_condition(self, condition_id, condition_context) -> bool:
        """Return if the condition exist and avilible"""
        if condition_id in self.conditions:
            condition = self.get_condition(condition_id)
            return condition.is_available(condition_context)
        logger.warning("Condition id %s not found.", condition_id)
        return False

    def get_no_param_condition(self) -> list:
        """Return the list of condition without parameters"""
        return self.no_parameter_conditions

    @staticmethod
    def check_and_sort(collections, target_type, reply):
        """Check the collection and sort conditions"""
        collection = collections.get(target_type)
        if collection:
            collection = sorted(collection, key=lambda x: x.get('id'))
            reply.append({"id": target_type + "_condition_collection", "conditions": collection})
        else:
            logger.warning("Condition collection for %s is None.", target_type)

    def get_all_collections(self, condition_context):
        """Get all register conditions."""

        collections = {
            TargetTypeEnum.WEIGHT.value: [], TargetTypeEnum.TENSOR.value: [], TargetTypeEnum.GRADIENT.value: [],
            TargetTypeEnum.ACTIVATION.value: []
        }
        for condition in self.conditions.values():
            parameters = []
            if not condition.is_available(condition_context):
                continue
            for param in condition.parameters:
                if not param.visible_on_ui:
                    continue
                parameters.append({
                    "name": param.name,
                    "type": param.type.name,
                    "support_disable": param.support_disable,
                    "default_value": param.default_value,
                    "param_type": param.param_type,
                    "required_params": param.required_params
                })
            collections[condition.supported_target_type.value].append({
                "id": condition.id,
                "parameters": parameters,
                "supported_target_type": condition.supported_target_type.name,
                "abbr": condition.abbr
            })

        reply = []
        self.check_and_sort(collections, TargetTypeEnum.TENSOR.value, reply)
        self.check_and_sort(collections, TargetTypeEnum.WEIGHT.value, reply)
        self.check_and_sort(collections, TargetTypeEnum.ACTIVATION.value, reply)
        self.check_and_sort(collections, TargetTypeEnum.GRADIENT.value, reply)

        return reply



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Compute resource manager."""


import functools
import multiprocessing
import threading
from concurrent import futures

import _thread

from mindinsight.utils.log import setup_logger

_MP_CONTEXT = multiprocessing.get_context(method="forkserver")


class ComputingResourceManager:
    """
    Manager for computing resources.

    Note:
        1. Please always use the get_instance method to get instance.
        2. This class will be used in a multi-threaded env, so it needs to be
           thread-safe.
    """
    _cls_lock = threading.Lock()
    _instance = None
    _exiting = False

    def __init__(self):
        self._executors = {}
        self._executor_id_counter = 1
        self._lock = threading.Lock()
        self._exiting = False
        self._logger = setup_logger("utils", "utils")

    @classmethod
    def get_instance(cls):
        """Get the singleton instance."""
        with cls._cls_lock:
            if cls._instance is None:
                cls._instance = ComputingResourceManager()
            return cls._instance

    def exit(self):
        """
        Called when the gunicorn worker process is exiting.

        This method will be called in the signal handling thread, which is not
        the same thread with get_executor. Also, this method will hold the lock
        to block other threads from operating the singleton or executors.
        """
        with self._lock:
            self._logger.info("Start to exit.")
            self._exiting = True
            for executor in self._executors.values():
                # It's safe to call executor.shutdown() multiple times.
                executor.shutdown(wait=True)
        self._logger.info("Exited.")

    def get_executor(self, max_processes_cnt=1):
        """
        Get an executor.

        This method may be called by different business from different threads.
        So it needs to be tread-safe.
        """
        with self._lock:
            if self._exiting:
                self._logger.info(
                    "System is exiting, will terminate the thread.")
                _thread.exit()

            executor = Executor(
                max_processes_cnt=max_processes_cnt,
                exit_callback=functools.partial(
                    self._remove_executor,
                    executor_id=self._executor_id_counter),
                exit_check_fn=self._check_exit
            )

            self._executors[self._executor_id_counter] = executor
            self._executor_id_counter += 1
            return executor

    def _remove_executor(self, executor_id):
        with self._lock:
            self._executors.pop(executor_id)

    def _check_exit(self):
        with self._lock:
            return self._exiting


class Executor:
    """
    Wrapped ProcessPoolExecutor to help global management.

    Args:
        max_processes_cnt (int): Max processes to use.
        exit_callback (Callable): A callback that will be called after process
            pool exit.
        exit_check_fn (Callable): A function to check whether the system is
            exiting.
    """
    def __init__(self, max_processes_cnt, exit_callback, exit_check_fn):
        self._backend = futures.ProcessPoolExecutor(
            max_workers=max_processes_cnt,
            mp_context=_MP_CONTEXT)
        self._exit_callback = exit_callback
        self._task_slots = threading.Semaphore(value=max_processes_cnt)
        self._exit_check_fn = exit_check_fn
        self._logger = setup_logger("utils", "utils")

    def __enter__(self):
        self._backend.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        ret = self._backend.__exit__(*args, **kwargs)
        self._exit_callback()
        return ret

    def submit(self, *args, **kwargs):
        if self._exit_check_fn():
            self._logger.warning(
                "System exiting, will terminate current thread.")
            _thread.exit()
        self._task_slots.acquire()
        future = self._backend.submit(*args, **kwargs)
        # The future object is not needed for releasing semaphores.
        future.add_done_callback(lambda future_obj: self._task_slots.release())
        return future

    submit.__doc__ = futures.Executor.submit.__doc__

    def shutdown(self, wait):
        self._backend.shutdown(wait)

    shutdown.__doc__ = futures.Executor.shutdown.__doc__


def terminate():
    """Set the terminating flag."""
    ComputingResourceManager.get_instance().exit()
